import io
import json
import logging
import tempfile
import webbrowser
from abc import ABC
from pathlib import Path
from typing import Optional

import omf
import pandas as pd
from omf import Project


class OMFPandasBase(ABC):

    def __init__(self, filepath: Path):
        """Instantiate the OMFPandas object.

        Args:
            filepath (Path): Path to the OMF file.

        Raises:
            FileNotFoundError: If the OMF file does not exist.
            ValueError: If the file is not an OMF file.
        """
        self._logger = logging.getLogger(__class__.__name__)
        if not filepath.suffix == '.omf':
            raise ValueError(f'File is not an OMF file: {filepath}')
        self.filepath: Path = filepath
        self.project: Optional[Project] = None
        if filepath.exists():
            self.project = omf.load(str(filepath))
        self._elements = self.project.elements if self.project else []
        self.elements: dict[str, str] = {e.name: e.__class__.__name__ for e in self._elements}
        self.element_attributes: dict[str, list[str]] = {e.name: [a.name for a in e.attributes] for e in self._elements}

    def __repr__(self):
        return f"OMFPandasBase({self.filepath})"

    def __str__(self):
        return f"OMFPandasBase object for OMF file: {self.filepath}"

    @property
    def changelog(self) -> Optional[pd.DataFrame]:
        """Return the change log as a DataFrame."""
        if 'changelog' not in self.project.metadata:
            return None
        return pd.DataFrame([json.loads(msg) for msg in self.project.metadata['changelog']])

    def get_element_by_name(self, element_name: str):
        """Get an element by its name.

        :param element_name: The name of the element to retrieve.
        :return:
        """
        element = [e for e in self._elements if e.name == element_name]
        if not element:
            raise ValueError(f"Element '{element_name}' not found in the OMF file: {self.filepath.name}. "
                             f"Available elements are: {list(self.elements.keys())}")
        elif len(element) > 1:
            raise ValueError(f"Multiple elements with the name '{element_name}' found in the OMF file: "
                             f"{self.filepath.name}")
        return element[0]

    def get_element_attribute_names(self, element_name: str) -> list[str]:
        """Get the attribute names of an element.

        :param element_name: The name of the element to retrieve.
        :return:
        """
        element = self.get_element_by_name(element_name)
        return [attr.name for attr in element.attributes]

    def view_block_model_profile(self, blockmodel_name: str, query: Optional[str] = None):
        """View the profile of a BlockModel in the default web browser.

        Args:
            blockmodel_name (str): The name of the BlockModel to profile.
            query (str): A query defining the subset of the BlockModel.
        """

        el = self.get_element_by_name(blockmodel_name)
        filter_key: str = query if query else 'no_filter'

        if el.metadata.get('profile') is None:
            raise ValueError(f"BlockModel '{blockmodel_name}' has not been profiled.  "
                             f"Please run 'profile_blockmodel' first.")

        # JSON string containing the profile report
        profile_html: str = el.metadata['profile'][filter_key]

        # Create a temporary file
        with tempfile.NamedTemporaryFile(delete=False, suffix=".html") as temp_file:
            temp_file.write(profile_html.encode('utf-8'))
            temp_file_path = temp_file.name

        # Open the temporary file in the default web browser
        webbrowser.open(f"file://{temp_file_path}")



import getpass
import json
from dataclasses import dataclass, field, asdict
from datetime import datetime
from typing import Literal


@dataclass
class ChangeMessage:
    """Dataclass to mange auditability"""
    element: str
    action: Literal['create', 'update', 'delete']
    description: str
    user: str = field(default_factory=getpass.getuser)
    timestamp: str = field(default_factory=datetime.now)

    def __str__(self):
        return json.dumps(asdict(self), default=str)




_pa = None
_yp = None


def _import_pandera():
    """Helper method to import pandera and handle ImportError."""
    global _pa
    if _pa is None:
        try:
            import pandera as pa
            _pa = pa
        except ImportError:
            raise ImportError("pandera is required to run this method. "
                              "Please install it by running 'poetry install omfpandas --extras validate' "
                              "or 'pip install pandera[io]'")
    return _pa


def _import_ydata_profiling():
    """Helper method to import ydata_profiling and handle ImportError."""
    global _yp
    if _yp is None:
        try:
            import ydata_profiling as yp
            _yp = yp
        except ImportError:
            raise ImportError("ydata-profiling is required to run this method. "
                              "Please install it by running 'poetry install omfpandas --extras profile' "
                              "or 'pip install ydata-profiling'")
    return _yp



import getpass
import json
from pathlib import Path
from typing import Optional, Literal

import omf
import pandas as pd
import ydata_profiling

from omfpandas import OMFPandasReader
from omfpandas.audit import ChangeMessage
from omfpandas.base import OMFPandasBase
from omfpandas.blockmodel import df_to_blockmodel, series_to_attribute

from omfpandas.extras import _import_ydata_profiling, _import_pandera
from omfpandas.utils.timer import log_timer


class OMFPandasWriter(OMFPandasBase):
    """A class to write pandas dataframes to an OMF file.

    Attributes:
        filepath (Path): Path to the OMF file.
    """

    def __init__(self, filepath: Path):
        """Instantiate the OMFPandasWriter object.

        Args:
            filepath (Path): Path to the OMF file.
        """
        super().__init__(filepath)
        self.user_id = getpass.getuser()

        if not filepath.exists():
            # log a message and create a new project
            project = omf.Project()
            project.name = filepath.stem
            project.description = f"OMF file created by OMFPandasWriter: {filepath.name}"
            self._logger.info(f"Creating new OMF file: {filepath}")
            self.project = project  # to enable the write_to_changelog method
            # create the audit record, which also saves the file
            self.write_to_changelog(element='None', action='create', description=f"File created: {filepath}")

        super().__init__(filepath)

    @log_timer()
    def write_blockmodel(self, blocks: pd.DataFrame, blockmodel_name: str, pd_schema_filepath: Optional[Path] = None,
                         allow_overwrite: bool = False):
        """Write a dataframe to a BlockModel.

        Only dataframes with centroid (x, y, z) and block dims (dx, dy, dz) indexes are supported.

        Args:
            blocks (pd.DataFrame): The dataframe to write to the BlockModel.
            blockmodel_name (str): The name of the BlockModel to write to.
            pd_schema_filepath (Optional[Path]): The path to the Pandera schema file. Default is None.  If provided,
                the schema will be used to validate the dataframe before writing.
            allow_overwrite (bool): If True, overwrite the existing BlockModel. Default is False.

        Raises:
            ValueError: If the element retrieved is not a BlockModel.
        """

        if pd_schema_filepath:
            pa = _import_pandera()
            # validate the dataframe, which may modify it via coercion
            pd_schema = pa.DataFrameSchema.from_yaml(pd_schema_filepath)
            self._logger.info(f"Validating dataframe with schema: {pd_schema_filepath}")
            blocks = pd_schema.validate(blocks)
            self._logger.info(f"Writing dataframe to BlockModel: {blockmodel_name}")
            bm = df_to_blockmodel(blocks, blockmodel_name)
            # persist the schema inside the omf file
            bm.metadata['pd_schema'] = pd_schema.to_json()
        else:
            self._logger.info(f"Writing dataframe to BlockModel: {blockmodel_name}")
            bm = df_to_blockmodel(blocks, blockmodel_name)

        if bm.name in [element.name for element in self.project.elements]:
            if not allow_overwrite:
                raise ValueError(f"BlockModel '{blockmodel_name}' already exists in the OMF file: {self.filepath}.  "
                                 f"If you want to overwrite, set allow_overwrite=True.")
            else:
                # remove the existing volume from the project
                volume_to_remove = [element for element in self.project.elements if element.name == bm.name][0]
                self.project.elements.remove(volume_to_remove)

        self.project.elements.append(bm)

        # create the audit record, which also saves the file
        self.write_to_changelog(element=bm.name, action='create', description='BlockModel written')

    def write_to_changelog(self, element: str, action: Literal['create', 'update', 'delete'], description: str):
        """Write a change message to the OMF file.

        Args:
            element: The name of the element that was changed
            action: The action taken on the object
            description: Description of the change

        Returns:

        """

        if 'changelog' not in self.project.metadata:
            self.project.metadata['changelog'] = []
        msg = ChangeMessage(element=element, user=self.user_id, action=action, description=description)
        self.project.metadata['changelog'].append(str(msg))
        omf.save(project=self.project, filename=str(self.filepath), mode='w')
        self.project = omf.load(str(self.filepath))

    def write_blockmodel_attribute(self, blockmodel_name: str, series: pd.Series,
                                   allow_overwrite: bool = False):
        """Write data to a specific attribute of a BlockModel.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            series (pd.Series): The data to write to the attribute.
            allow_overwrite (bool): If True, overwrite the existing attribute. Default is False.
        """

        bm = self.get_element_by_name(blockmodel_name)
        if bm.metadata.get('pd_schema'):
            pa = _import_pandera()
            # validate the data
            schema = pa.io.from_json(bm.metadata['pd_schema'])
            series = schema.validate(series.to_frame())
            series = series.iloc[:, 0]  # back to series

        attrs: list[str] = self.get_element_attribute_names(blockmodel_name)
        if series.name in attrs:
            if allow_overwrite:
                # get the index in the list
                attr_pos = attrs.index(str(series.name))
                bm.attributes[attr_pos] = series_to_attribute(series)
            else:
                raise ValueError(f"Attribute '{series.name}' already exists in BlockModel '{blockmodel_name}'.  "
                                 f"If you want to overwrite, set allow_overwrite=True.")
        else:
            bm.attributes.append(series_to_attribute(series))

        self._delete_profile_report(blockmodel_name)

        # todo: re-profile...

        self.write_to_changelog(element=bm.name, action='create', description=f"Attribute [{series.name}] written")

    def delete_blockmodel_attribute(self, blockmodel_name: str, attribute_name: str):
        """Delete an attribute from a BlockModel.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            attribute_name (str): The name of the attribute.
        """
        bm = self.get_element_by_name(blockmodel_name)
        attrs: list[str] = self.get_element_attribute_names(bm)
        if attribute_name in attrs:
            del bm.attributes[attribute_name]
        else:
            raise ValueError(f"Attribute '{attribute_name}' not found in BlockModel '{blockmodel_name}'.")

        self._delete_profile_report(blockmodel_name)

        # create the audit record, which also saves the file
        self.write_to_changelog(element=bm.name, action='delete', description=f"{attribute_name} deleted")

    @log_timer()
    def profile_blockmodel(self, blockmodel_name: str, query: Optional[str] = None):
        """Profile a BlockModel.

        Profiling will be skipped if the data has not changed.

        Args:
            blockmodel_name (str): The name of the BlockModel to profile.
            query (Optional[str]): A query to filter the data before profiling.

        Returns:
            pd.DataFrame: The profiled data.
        """

        _import_ydata_profiling()

        df: pd.DataFrame = OMFPandasReader(self.filepath).read_blockmodel(blockmodel_name, query=query)
        el = self.get_element_by_name(blockmodel_name)
        bm_type = str(type(el)).split('.')[-1].rstrip("'>")
        dataset: dict = {"description": f"{el.description} Filter: {query if query else 'no_filter'}",
                         "creator": self.user_id, "url": self.filepath.as_uri()}
        column_descriptions: dict = {}
        if el.metadata.get('pd_schema'):
            column_defs: dict = json.loads(el.metadata['pd_schema'])['columns']
            column_descriptions = {k: f"{v['title']}: {v['description']}" for k, v in column_defs.items()}

        profile: ydata_profiling.ProfileReport = df.profile_report(title=f"{el.name} {bm_type}", dataset=dataset,
                                                                   variables={"descriptions": column_descriptions})

        # persist the profile report as html to the omf file, larger but cannot serialise the ProfileReport object,
        # nor recreate the report from json.
        d_profile: dict = {query if query else 'no_filter': profile.to_html()}

        if el.metadata.get('profile'):
            el.metadata['profile'] = {**el.metadata['profile'], **d_profile}
        else:
            el.metadata['profile'] = d_profile

        self.write_to_changelog(element=blockmodel_name, action='create', description=f"Profiled with query {query}")

        return profile

    def write_block_model_schema(self, blockmodel_name: str, pd_schema_filepath: Path):
        """Write a Pandera schema to the OMF file.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            pd_schema_filepath (Path): The path to the Pandera schema yaml file.
        """
        pa = _import_pandera()
        bm = self.get_element_by_name(blockmodel_name)
        pd_schema = pa.DataFrameSchema.from_yaml(pd_schema_filepath)
        bm.metadata['pd_schema'] = pd_schema.to_json()

        el = self.get_element_by_name(blockmodel_name)
        schema_title = pd_schema.title if pd_schema.title else ''
        schema_description = pd_schema.description if pd_schema.description else ''
        el.description = f"{schema_title}: {schema_description}"

        self.write_to_changelog(element=blockmodel_name, action='create', description=f"Schema written")

    def _delete_profile_report(self, blockmodel_name: str):
        """Delete the profile report from the OMF file when data has changed."""
        bm = self.get_element_by_name(blockmodel_name)

        if 'profile' in bm.metadata:
            del bm.metadata['profile']




from pathlib import Path

from omfpandas.base import OMFPandasBase
from omfpandas.blockmodel import blockmodel_to_parquet
from omfpandas.utils.timer import log_timer


class OMFDataConverter(OMFPandasBase):
    """A class to handle conversions between OMF and other formats."""

    def __init__(self, filepath: Path):
        """Instantiate the OMFConverter object

        Args:
            filepath: Path to the OMF file.
        """
        if not filepath.exists():
            raise FileNotFoundError(f'File does not exist: {filepath}')
        super().__init__(filepath)

    @log_timer()
    def blockmodel_to_parquet(self, blockmodel_name: str, parquet_filepath: Path,
                              allow_overwrite: bool = False):
        """Write a VolumeElement to a Parquet file.

        Args:
            blockmodel_name (str): The name of the VolumeElement to convert.
            parquet_filepath (Path): The path to the Parquet file to write.
            allow_overwrite (bool): If True, overwrite the existing Parquet file. Default is False.

        Raises:
            ValueError: If the element retrieved is not a VolumeElement.
        """
        bm = self.get_element_by_name(blockmodel_name)
        if bm.__class__.__name__ not in ['RegularBlockModel', 'TensorGridBlockModel']:
            raise ValueError(
                f"Element '{blockmodel_name}' is not a supported BlockModel in the OMF file: {self.filepath}")

        blockmodel_to_parquet(blockmodel=bm, out_path=parquet_filepath,
                              allow_overwrite=allow_overwrite)



from omfpandas.reader import OMFPandasReader
from omfpandas.writer import OMFPandasWriter
from omfpandas.converter import OMFDataConverter

from importlib import metadata

try:
    __version__ = metadata.version('omfpandas')
except metadata.PackageNotFoundError:
    # Package is not installed
    pass



from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd

from omfpandas.base import OMFPandasBase
from omfpandas.blockmodel import blockmodel_to_df, create_index
from omfpandas.utils.pandas import parse_vars_from_expr


class OMFPandasReader(OMFPandasBase):
    """A class to read an OMF file to a pandas DataFrame.

    Attributes:
        filepath (Path): Path to the OMF file.

    """

    def __init__(self, filepath: Path):
        """Instantiate the OMFPandasReader object

        Args:
            filepath: Path to the OMF file.
        """
        if not filepath.exists():
            raise FileNotFoundError(f'File does not exist: {filepath}')
        super().__init__(filepath)

    def read_blockmodel(self, blockmodel_name: str, attributes: Optional[list[str]] = None,
                        query: Optional[str] = None, index_filter: Optional[list[int]] = None) -> pd.DataFrame:
        """Return a DataFrame from a BlockModel.

        Only variables assigned to the `cell` (as distinct from the grid `points`) are loaded.

        Args:
            blockmodel_name (str): The name of the BlockModel to convert.
            attributes (Optional[list[str]]): The attributes/variables to include in the DataFrame. If None, all
                variables are included.
            query (Optional[str]): A query string to filter the DataFrame. Default is None.
            index_filter (Optional[list[int]]): A list of indexes to filter the DataFrame. Default is None.

        Returns:
            pd.DataFrame: The DataFrame representing the BlockModel.
        """
        bm = self.get_element_by_name(blockmodel_name)
        # check the element retrieved is the expected type
        if bm.__class__.__name__ not in ['RegularBlockModel', 'TensorGridBlockModel']:
            raise ValueError(f"Element '{bm}' is not a supported BlockModel in the OMF file: {self.filepath}")

        return blockmodel_to_df(bm, variables=attributes, query=query, index_filter=index_filter)

    def read_block_models(self, blockmodel_attributes: dict[str, list[str]],
                          query: Optional[str] = None) -> pd.DataFrame:
        """Return a DataFrame from multiple BlockModels.

        Args:
            blockmodel_attributes (dict[str, list[str]]): A dictionary of BlockModel names and the variables to include.
                If the dict value is None, all attributes in the blockmodel (key) are included.
            query (Optional[str]): A query string to filter the DataFrame. Default is None.

        Returns:
            pd.DataFrame: The DataFrame representing the merged BlockModels.
        """
        block_models: dict[str, pd.DataFrame] = {}
        geometry_indexes: dict[str, pd.MultiIndex] = {}
        index_filter = None
        if query:
            # we use the query to generate an index_filter to filter the block models since they may not have
            # the query attributes available in the given model.
            query_attrs: list[str] = parse_vars_from_expr(query)
            # extract the attributes for the query_attributes, noting that they may be in different block models
            chunks: list[pd.Series] = []
            for bm_name in blockmodel_attributes:
                for attr in query_attrs:
                    if attr in self.element_attributes[bm_name]:
                        chunks.append(self.read_blockmodel(blockmodel_name=bm_name, attributes=[attr])[attr])
            tmp_df: pd.DataFrame = pd.concat(chunks, axis=1).query(query)
            # get the index locations of the filtered index relative to the full index
            filtered_index = tmp_df.index
            full_index = create_index(self.get_element_by_name(self._elements[0].name))
            # Find the intersection of the two MultiIndex objects
            intersection = full_index.intersection(filtered_index)
            index_filter = full_index.get_indexer(intersection)

        for bm_name, requested_attrs in blockmodel_attributes.items():
            # check that the requested attrs exist in the specified bm
            available_attrs = self.element_attributes[bm_name]
            if requested_attrs is None:
                requested_attrs = available_attrs
            else:
                missing_attrs = set(requested_attrs) - set(available_attrs)
                if missing_attrs:
                    raise ValueError(f"Attributes {missing_attrs} not found in BlockModel '{bm_name}'. "
                                     f"Available attributes are: {available_attrs}")

            block_models[bm_name] = self.read_blockmodel(blockmodel_name=bm_name, attributes=requested_attrs,
                                                         index_filter=index_filter if query else None)
            geometry_indexes[bm_name] = create_index(blockmodel=self.get_element_by_name(bm_name))

        # validate the indexes are equivalent
        def ensure_identical_indexes(index_dict: dict[str, pd.MultiIndex]) -> None:
            if not index_dict:
                return

            first_index = next(iter(index_dict.values()))
            for name, index in index_dict.items():
                if not first_index.equals(index):
                    raise ValueError(f"Index for '{name}' is different from the first index.")

        ensure_identical_indexes(geometry_indexes)

        return pd.concat(block_models.values(), axis=1)



import tokenize
from io import StringIO
from token import STRING

import pandas as pd


def is_nullable_integer_dtype(series: pd.Series) -> bool:
    """

    Args:
        series: The series

    Returns:
        bool: True if series contains nullable integer
    """

    return True if str(series.dtype)[0] == "I" else False


def to_nullable_integer_dtype(series: pd.Series) -> pd.Series:
    """ Convert an int series to a nullable integer dtype

    Args:
        series: The series

    Returns:
        pd.Series: The series with nullable dtype
    """

    return series.astype(str(series.dtype).replace("i", "I")) if is_nullable_integer_dtype(series) else series


def to_numpy_integer_dtype(series: pd.Series) -> pd.Series:
    """ Convert a nullable int series to a numpy integer dtype

    Args:
        series: The series

    Returns:
        pd.Series: The series with nullable dtype
    """

    return series.astype(str(series.dtype).replace("I", "i")) if is_nullable_integer_dtype(series) else series


def parse_vars_from_expr(expr: str) -> list[str]:
    """ Parse variables from a pandas query expression string.

    Args:
        expr: The expression string

    Returns:
        list[str]: The list of variables
    """
    variables = set()
    tokens = tokenize.generate_tokens(StringIO(expr).readline)
    logical_operators = {'and', 'or', '&', '|'}
    inside_backticks = False
    current_var = []

    for token in tokens:
        if token.string == '`':
            if inside_backticks:
                # End of backtick-enclosed variable
                variables.add(' '.join(current_var))
                current_var = []
            inside_backticks = not inside_backticks
        elif inside_backticks:
            if token.type in {tokenize.NAME, STRING}:
                current_var.append(token.string)
        elif token.type == tokenize.NAME and token.string not in logical_operators:
            variables.add(token.string)

    return list(variables)



import ast
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, TypeVar, Union

import numpy as np
import pandas as pd
from omf import NumericAttribute, CategoryAttribute, CategoryColormap

from omf.blockmodel import BaseBlockModel, RegularBlockModel, TensorGridBlockModel
from pandas.core.dtypes.common import is_integer_dtype

from omfpandas.utils.pandas import is_nullable_integer_dtype, to_nullable_integer_dtype, to_numpy_integer_dtype, \
    parse_vars_from_expr

# generic type variable, used for type hinting, to indicate that the type is a subclass of BaseBlockModel
BM = TypeVar('BM', bound=BaseBlockModel)

SENTINEL_VALUE = -9  # TODO: possibly move to config file


@dataclass
class TensorGeometry:
    """A dataclass to represent the geometry of a tensor grid block model."""
    origin: np.ndarray
    axis_u: np.ndarray
    axis_v: np.ndarray
    axis_w: np.ndarray
    tensor_u: np.ndarray
    tensor_v: np.ndarray
    tensor_w: np.ndarray

    def is_regular(self) -> bool:
        """Return True if the tensor grid is regular."""
        return (np.allclose(self.tensor_u, self.tensor_u[0]) and np.allclose(self.tensor_v, self.tensor_v[0]) and
                np.allclose(self.tensor_w, self.tensor_w[0]))


def blockmodel_to_df(blockmodel: BM, variables: Optional[list[str]] = None,
                     query: Optional[str] = None,
                     index_filter: Optional[list[int]] = None) -> pd.DataFrame:
    """Convert block model to a DataFrame.

    Args:
        blockmodel (BlockModel): The BlockModel to convert.
        variables (Optional[list[str]]): The variables to include in the DataFrame. If None, all variables are included.
        query (Optional[str]): The query to filter the DataFrame.
        index_filter (Optional[list[int]]): List of integer indices to filter the DataFrame.

    Returns:
        pd.DataFrame: The DataFrame representing the BlockModel.
    """
    # read the data
    df: pd.DataFrame = read_blockmodel_attributes(blockmodel, attributes=variables, query=query, index_filter=index_filter)
    return df


def df_to_blockmodel(df: pd.DataFrame, blockmodel_name: str, is_tensor: bool = True) -> BM:
    """Write a DataFrame to a BlockModel.

    Args:
        df (pd.DataFrame): The DataFrame to convert to a BlockModel.
        blockmodel_name (str): The name of the BlockModel.
        is_tensor (bool): If True, a TensorGridBlockModel will be created. If False, a RegularBlockModel will be
        created.

    Returns:
        BlockModel: The BlockModel representing the DataFrame.
    """
    # Get the original order
    original_order = df.index

    try:
        # Sort the dataframe to align with the omf spec
        df.sort_index(level=['z', 'y', 'x'], inplace=True)

        # Create the blockmodel and geometry
        geometry: TensorGeometry = index_to_geometry(df.index)

        if is_tensor:
            blockmodel: BM = TensorGridBlockModel(name=blockmodel_name)
            # assign the geometry properties
            blockmodel.corner = geometry.origin
            blockmodel.axis_u = geometry.axis_u
            blockmodel.axis_v = geometry.axis_v
            blockmodel.axis_w = geometry.axis_w
            blockmodel.tensor_u = geometry.tensor_u
            blockmodel.tensor_v = geometry.tensor_v
            blockmodel.tensor_w = geometry.tensor_w
        else:
            if not geometry.is_regular():
                raise ValueError("RegularBlockModel requires a regular grid.")
            blockmodel: BM = RegularBlockModel(name=blockmodel_name)
            blockmodel.corner = geometry.origin
            blockmodel.axis_u = geometry.axis_u
            blockmodel.axis_v = geometry.axis_v
            blockmodel.axis_w = geometry.axis_w
            blockmodel.block_count = np.ndarray(
                [geometry.tensor_u.size, geometry.tensor_v.size, geometry.tensor_w.size])
            blockmodel.block_size = np.ndarray([geometry.tensor_u[0], geometry.tensor_v[0], geometry.tensor_w[0]])

        # add the data
        attrs: list[Union[NumericAttribute, CategoryAttribute]] = []
        for variable in df.columns:
            attribute = series_to_attribute(df[variable])

            attrs.append(attribute)
        blockmodel.attributes = attrs

    finally:
        # Reset the index to the original order (to avoid side effects)
        df = df.reindex(original_order)

    return blockmodel


def series_to_attribute(series: pd.Series) -> Union[CategoryAttribute, NumericAttribute]:
    if isinstance(series.dtype, pd.CategoricalDtype):
        cat_map = {i: c for i, c in enumerate(series.cat.categories)}
        cat_col_map = CategoryColormap(indices=list(cat_map.keys()), values=list(cat_map.values()))
        attribute = CategoryAttribute(name=series.name, location="cells", array=np.array(series.cat.codes),
                                      categories=cat_col_map)
    else:
        # manage the sentinel / null placeholders
        # REF: https://github.com/gmggroup/omf-python/issues/59
        if is_nullable_integer_dtype(series):
            # set null_values and assign metadata
            data: pd.Series = series.fillna(SENTINEL_VALUE).pipe(to_numpy_integer_dtype)
            attribute = NumericAttribute(name=series.name, location="cells", array=data.values)
            attribute.metadata['null_value'] = SENTINEL_VALUE
        elif is_integer_dtype(series):
            attribute = NumericAttribute(name=series.name, location="cells", array=series.values)
            attribute.metadata['null_value'] = SENTINEL_VALUE
        else:
            attribute = NumericAttribute(name=series.name, location="cells", array=series.values)
            attribute.metadata['null_value'] = 'np.nan'
    return attribute


def attribute_to_series(attribute: Union[CategoryAttribute, NumericAttribute]) -> pd.Series:
    if isinstance(attribute, CategoryAttribute):
        return pd.Series(pd.Categorical.from_codes(codes=attribute.array.array.ravel(),
                                                   categories=attribute.categories.values,
                                                   ordered=False), name=attribute.name)
    else:
        # if an int with null_value in metadata then convert to a nullable int
        if attribute.metadata.get("null_value") and is_integer_dtype(attribute.array.array):
            return pd.Series(attribute.array.array.ravel(), name=attribute.name).pipe(
                to_nullable_integer_dtype).replace(
                SENTINEL_VALUE, pd.NA)
        return pd.Series(attribute.array.array.ravel(), name=attribute.name, dtype=attribute.array.array.dtype)


def blockmodel_to_parquet(blockmodel: BM, out_path: Optional[Path] = None,
                          variables: Optional[list[str]] = None,
                          allow_overwrite: bool = False):
    """Convert blockmodel to a Parquet file.

    Args:
        blockmodel (BlockModel): The BlockModel to convert.
        out_path (Optional[Path]): The path to the Parquet file to write. If None, a file with the blockmodel name is
        created.
        variables (Optional[list[str]]): The variables to include in the DataFrame. If None, all variables are included.
        allow_overwrite (bool): If True, overwrite the existing Parquet file. Default is False.

    Raises:
        FileExistsError: If the file already exists and allow_overwrite is False.
    """
    if out_path is None:
        out_path = Path(f"{blockmodel.name}.parquet")
    if out_path.exists() and not allow_overwrite:
        raise FileExistsError(f"File already exists: {out_path}. If you want to overwrite, set allow_overwrite=True.")
    df: pd.DataFrame = blockmodel_to_df(blockmodel, variables=variables)
    df.to_parquet(out_path)


def read_blockmodel_attributes(blockmodel: BM, attributes: Optional[list[str]] = None,
                               query: Optional[str] = None, index_filter: Optional[list[int]] = None) -> pd.DataFrame:
    """Read the attributes/variables from the BlockModel.

    Args:
        blockmodel (BlockModel): The BlockModel to read from.
        attributes (list[str]): The attributes to include in the DataFrame.
        query (str): The query to filter the DataFrame.
        index_filter (list[int]): List of integer indices to filter the DataFrame.

    Returns:
        pd.DataFrame: The DataFrame representing the attributes in the BlockModel.

    Raises:
        ValueError: If the attribute is not found in the BlockModel or if both query and index_filter are provided.
    """
    if query and index_filter:
        raise ValueError("Cannot use both query and index_filter at the same time.")

    # identify 'cell' variables in the file
    attributes_available = [v.name for v in blockmodel.attributes if v.location == 'cells']

    attributes = attributes or attributes_available

    # check if the variables are available
    if not set(attributes).issubset(attributes_available):
        raise ValueError(f"Variables {set(attributes).difference(attributes_available)} not found in the BlockModel.")

    int_index: np.ndarray = np.arange(blockmodel.num_cells)
    if query is not None:
        # parse out the attributes from the query using a package
        query_attrs = parse_vars_from_expr(query)
        # check if the attributes in the query are available
        if not set(query_attrs).issubset(attributes_available):
            raise ValueError(
                f"Variables {set(query_attrs).difference(attributes_available)} not found in the BlockModel.")
        query_series: list = []
        for attr_name in query_attrs:
            query_series.append(attribute_to_series(_get_attribute_by_name(blockmodel, attr_name)))
        df_to_query: pd.DataFrame = pd.concat(query_series, axis=1)
        int_index = np.array(df_to_query.query(query).index)
    elif index_filter is not None:
        int_index = np.array(index_filter)

    # Loop over the variables
    chunks: list = []
    for attr in attributes:
        attr: Union[CategoryAttribute, NumericAttribute] = _get_attribute_by_name(blockmodel, attr)
        chunks.append(attribute_to_series(attr).iloc[int_index])

    # create the geometry index
    geometry_index = create_index(blockmodel)
    if (query is not None) or (index_filter is not None):
        # filter the index to match the int_index positional index
        geometry_index = geometry_index.take(int_index)

    res = pd.concat(chunks, axis=1)
    res.index = geometry_index
    return res if isinstance(res, pd.DataFrame) else res.to_frame()


def create_index(blockmodel: BM) -> pd.MultiIndex:
    """Returns a pd.MultiIndex for the blockmodel element.

    Args:
        blockmodel (BlockModel): The BlockModel to get the index from.

    Returns:
        pd.MultiIndex: The MultiIndex representing the blockmodel element geometry.
    """
    ox, oy, oz = blockmodel.corner

    # Make coordinates (points) along each axis, i, j, k
    i = ox + np.cumsum(blockmodel.tensor_u)
    i = np.insert(i, 0, ox)
    j = oy + np.cumsum(blockmodel.tensor_v)
    j = np.insert(j, 0, oy)
    k = oz + np.cumsum(blockmodel.tensor_w)
    k = np.insert(k, 0, oz)

    # convert to centroids
    x, y, z = (i[1:] + i[:-1]) / 2, (j[1:] + j[:-1]) / 2, (k[1:] + k[:-1]) / 2
    xx, yy, zz = np.meshgrid(x, y, z, indexing="ij")

    # Calculate dx, dy, dz
    dxx, dyy, dzz = np.meshgrid(blockmodel.tensor_u, blockmodel.tensor_v, blockmodel.tensor_w, indexing="ij")

    # TODO: consider rotation

    index = pd.MultiIndex.from_arrays([xx.ravel("F"), yy.ravel("F"), zz.ravel("F"),
                                       dxx.ravel("F"), dyy.ravel("F"), dzz.ravel("F")],
                                      names=['x', 'y', 'z', 'dx', 'dy', 'dz'])

    return index


def index_to_geometry(index: pd.MultiIndex) -> TensorGeometry:
    """Convert a MultiIndex to a VolumeGridGeometry.

    Args:
        index (pd.MultiIndex): The MultiIndex to convert to a TensorGeometry.

    Returns:
        TensorGeometry: The TensorGeometry representing the MultiIndex.
    """
    # check that the index contains the expected levels
    if not {'x', 'y', 'z', 'dx', 'dy', 'dz'}.issubset(index.names):
        raise ValueError("Index must contain the levels 'x', 'y', 'z', 'dx', 'dy', 'dz'.")

    x = index.get_level_values('x').unique()
    y = index.get_level_values('y').unique()
    z = index.get_level_values('z').unique()

    # Get the shape of the original 3D arrays
    shape = (len(x), len(y), len(z))

    # Reshape the ravelled index back into the original shapes
    tensor_u = index.get_level_values('dx').values.reshape(shape, order='F')[:, 0, 0]
    tensor_v = index.get_level_values('dy').values.reshape(shape, order='F')[0, :, 0]
    tensor_w = index.get_level_values('dz').values.reshape(shape, order='F')[0, 0, :]

    origin_x = x.min() - tensor_u[0] / 2
    origin_y = y.min() - tensor_v[0] / 2
    origin_z = z.min() - tensor_w[0] / 2

    # Create the geometry
    origin = np.array([origin_x, origin_y, origin_z])
    axis_u = np.array([1, 0, 0])
    axis_v = np.array([0, 1, 0])
    axis_w = np.array([0, 0, 1])
    geometry: TensorGeometry = TensorGeometry(origin=origin, axis_u=axis_u, axis_v=axis_v, axis_w=axis_w,
                                              tensor_u=tensor_u, tensor_v=tensor_v, tensor_w=tensor_w)

    return geometry


def _get_attribute_by_name(blockmodel: BM, attr_name: str) -> Union[CategoryAttribute, NumericAttribute]:
    """Get the variable/attribute by its name from a BlockModel.

    Args:
        blockmodel (BlockModel): The BlockModel to get the data from.
        attr_name (str): The name of the attribute to retrieve.

    Returns:
        Union[CategoryAttribute, NumericAttribute]: The attribute with the given name.

    Raises:
        ValueError: If the variable is not found as cell data in the BlockModel or if multiple variables with the
        same name are found.
    """
    attrs = [sd for sd in blockmodel.attributes if sd.location == 'cells' and sd.name == attr_name]
    if not attrs:
        raise ValueError(f"Variable '{attr_name}' not found as cell data in the BlockModel: {blockmodel}")
    elif len(attrs) > 1:
        raise ValueError(f"Multiple variables with the name '{attr_name}' found in the BlockModel: {blockmodel}")
    return attrs[0]
